/*
 * Factory.h
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#ifndef FACTORY_H_
#define FACTORY_H_

#include "IsoGaussian.h"
#include "GP.h"

class Factory
{
	private:

	Mat Field; // the entire field
	int N, nL; // the size of the field and the number of kernels

	vector <IsoGaussian*> covfuncs; // the set of kernels
	vector <GP*> models; // the set of corresponding models;
	Vec b; // current belief over possible kernels

	double offset; // submodularity constant for differential entropy w.r.t mixed GP models
	bool *marker; // check whether a location has been previously selected

	Vec X; // the index of selected locations
	int m; // the number of selected locations;
	vector <Mat> K, L; // current correlation matrix and its Cholesky decomposition

	Vec pX, pY; // prior data

	void Initialize(Vec &X, int k);
	double Entropy(int idx, int k);
	pair <double, int> NextLocation(int k);

	void Initialize(Vec &X);
	double Entropy(int idx);
	pair <double, int> NextLocation();

	public:

	Factory(Mat &Field, int N, vector <IsoGaussian*> &covfuncs, int nL);
	pair <double, Vec> GreedyDesign(Vec &X, int n, int k);
	pair <double, Vec> GreedyDesign(Vec &X, int n, Vec &b);

	pair <double, double> InitialBounds(Vec &X, Vec &b, int n);
	double SubmodularConstant();
	vector < vector < pair <double, double> > > PosteriorGP(Vec &X, Vec &Y);
	vector < pair <int, Vec> > Sample(Vec &X, vector < vector < pair <double, double> > > &pGP, Vec &b, int nSample);

	void GetPriorData(Vec &X, Vec &Y);
	void UpdatePriorData(int x, double y);
	vector < pair <double, double> > PosteriorGP(Vec &X, Vec &Y, int x);
};

#endif /* FACTORY_H_ */
