/*
 * IELearner.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "IELearner.h"

IELearner::IELearner(Mat &Field, int N, vector <IsoGaussian*> &covfunc, int nL, Vec &b)
{
	this->Field = Field; this->N = N; this->nL = nL;
	this->covfunc = covfunc; this->b = b;

	marker = new bool[N];
	for (int i = 0; i < N; i++) marker[i] = false;

	this->m = 0; this->X.clear(); this->Y.clear();
	this->K = vector <Mat> (this->nL); this->L = vector <Mat>(this->nL);

	this->offset = INFTY; this->MLB = -INFTY;

	for (int t = 0; t < this->nL; t++)
	{
		this->K[t].clear(); this->L[t].clear();
		this->offset = min(this->offset, this->covfunc[t]->Noise());
	}

	this->offset = -min(0.0, ENT(pow(this->offset, 2.0)));
}

void IELearner::Refresh(Vec &b)
{
	this->b = b; this->MLB = -INFTY;

	for (int i = 0; i < this->N; i++) this->marker[i] = false;

	this->m = 0; this->X.clear(); this->Y.clear();

	for (int t = 0; t < this->nL; t++)
	{
		this->K[t].clear(); this->L[t].clear();
	}
}

double IELearner::Entropy(int s, int kernel)
{
	double var = this->covfunc[kernel]->Var(this->Field[s]);

	if (this->m > 0)
	{
		Vec k(this->m, 0.0);
		for (int j = 0; j < this->m; j++) k[j] = this->covfunc[kernel]->Eval(this->Field[s], this->Field[X[j]]);
		Vec v = ForwardSub(this->L[kernel], k);
		var = var - Dot(v, v);
	}

	return ENT(var);
}

double IELearner::Entropy(int s)
{
	double res = 0.0;

	for (int t = 0; t < this->nL; t++)
	{
		double var = this->covfunc[t]->Var(this->Field[s]);

		if (this->m > 0)
		{
			Vec k(this->m, 0.0);
			for (int j = 0; j < this->m; j++) k[j] = this->covfunc[t]->Eval(this->Field[s], this->Field[X[j]]);
			Vec v = ForwardSub(this->L[t], k);
			var = var - Dot(v, v);
		}

		res = res + this->b[t] * ENT(var);
	}

	return res;
}

pair <double, int> IELearner::GreedyChoice(int kernel)
{
	pair <double, int> res = make_pair(-INFTY, -1);

	for (int s = 0; s < this->N; s++)
		if (!this->marker[s])
		{
			double entropy = Entropy(s, kernel);
			if (entropy > res.first)
			{
				res.first = entropy;
				res.second = s;
			}
		}

	return res;
}

pair <double, int> IELearner::GreedyChoice()
{
	pair <double, int> res = make_pair(-INFTY, -1);

	for (int s = 0; s < this->N; s++)
		if (!this->marker[s])
		{
			double entropy = Entropy(s);
			if (entropy > res.first)
			{
				res.first = entropy;
				res.second = s;
			}
		}

	return res;
}

pair <double, Vec> IELearner::AprioriDesign(int n, int kernel)
{
	pair <double, Vec> res = make_pair(0.0, Vec(n, -1));

	n = min(n, this->N - this->m);

	Mat Ks = this->K[kernel], Ls = this->L[kernel];

	for (int i = 0; i < n; i++)
	{
		pair <double, int> next = GreedyChoice(kernel);
		res.first += next.first;
		res.second[i] = next.second;

		this->marker[next.second] = true; this->X.push_back(next.second);
		for (int j = 0; j < this->m; j++) this->K[kernel][j].push_back(this->covfunc[kernel]->Eval(this->Field[next.second], this->Field[X[j]]));
		this->K[kernel].push_back(Vec(this->m + 1, 0.0));
		for (int j = 0; j < this->m; j++) this->K[kernel][this->m][j] = this->K[kernel][j][this->m];
		this->K[kernel][this->m][this->m] = this->covfunc[kernel]->Var(this->Field[next.second]); this->m = this->m + 1;

		this->L[kernel] = Cholesky(this->K[kernel]);
	}

	this->K[kernel] = Ks; this->L[kernel] = Ls;
	this->X.resize(m - n); this->m = this->m - n;
	for (int i = 0; i < n; i++) this->marker[(int)res.second[i]] = false;

	return res;
}

pair <double, Vec> IELearner::AprioriDesign(int n)
{
	pair <double, Vec> res = make_pair(0.0, Vec(n, -1));

	n = min(n, this->N - this->m);

	vector <Mat> Ks(this->nL); vector <Mat> Ls(this->nL);
	for (int t = 0; t < nL; t++)
	{
		Ks[t] = this->K[t];
		Ls[t] = this->L[t];
	}

	for (int i = 0; i < n; i++)
	{
		pair <double, int> next = GreedyChoice();
		res.first += next.first;
		res.second[i] = next.second;

		this->marker[next.second] = true; X.push_back(next.second);

		for (int t = 0; t < this->nL; t++)
		{
			for (int j = 0; j < this->m; j++) this->K[t][j].push_back(this->covfunc[t]->Eval(this->Field[next.second], this->Field[X[j]]));
			this->K[t].push_back(Vec(m + 1, 0.0));
			for (int j = 0; j < this->m; j++) this->K[t][this->m][j] = this->K[t][j][this->m]; this->K[t][m][m] = this->covfunc[t]->Var(this->Field[next.second]);
			this->L[t] = Cholesky(this->K[t]);
		}

		this->m = this->m + 1;
	}

	this->K = Ks; this->L = Ls;
	X.resize(this->m - n); this->m = this->m - n;
	for (int i = 0; i < n; i++) this->marker[(int) res.second[i]] = false;

	return res;
}

double IELearner::Likelihood(int selected, double obs, int t)
{
	double var = this->covfunc[t]->Var(this->Field[selected]), mean = 0.0;

	if (this->m > 0)
	{
		Vec k(this->m, 0.0);
		for (int j = 0; j < this->m; j++) k[j] = this->covfunc[t]->Eval(Field[selected], Field[X[j]]);
		Vec v = ForwardSub(this->L[t], k);
		var = var - Dot(v, v);

		Mat Lt = Trans(this->L[t]);
		Vec beta = ForwardSub(this->L[t], this->Y), alpha = BackwardSub(Lt, beta);
		mean = Dot(k, alpha);
	}

	return N(obs, mean, var);
}

void IELearner::LightUpdate(int selected, double obs)
{
	double sum = 0.0;

	for (int t = 0; t < this->nL; t++)
	{
		this->b[t] = this->b[t] * Likelihood(selected, obs, t);
		sum = sum + this->b[t];
	}

	for (int t = 0; t < this->nL; t++) this->b[t] = this->b[t] / sum;
}

void IELearner::Update(int selected, double obs)
{
	double sum = 0.0;

	for (int t = 0; t < this->nL; t++)
	{
		this->b[t] = this->b[t] * Likelihood(selected, obs, t);
		sum = sum + this->b[t];
	}

	for (int t = 0; t < this->nL; t++) this->b[t] = this->b[t] / sum;

	this->X.push_back(selected); this->Y.push_back(obs); this->marker[selected] = true;

	for (int t = 0; t < this->nL; t++)
	{
		for (int j = 0; j < this->m; j++) this->K[t][j].push_back(this->covfunc[t]->Eval(this->Field[selected], this->Field[X[j]]));
		this->K[t].push_back(Vec(this->m + 1, 0.0));
		for (int j = 0; j < this->m; j++) this->K[t][this->m][j] = this->K[t][j][this->m];
		this->K[t][this->m][this->m] = this->covfunc[t]->Var(this->Field[selected]);
		this->L[t] = Cholesky(this->K[t]);
	}

	this->m = this->m + 1;
}

Vec IELearner::Move(int n, double stop)
{
	n = min(n, this->N - this->m);

	pair <double, Vec> lb = AprioriDesign(n);

	vector < pair <double, Vec> > ubs(nL);

	double LB = lb.first, UB = 0.0;

	this->MLB = max(this->MLB, LB);

	for (int t = 0; t < this->nL; t++)
		UB -= this->b[t] * log(this->b[t]);

	for (int t = 0; t < this->nL; t++)
	{
		ubs[t] = AprioriDesign(n, t);
		UB += this->b[t] * ((E * ubs[t].first + n * this->offset) / (E - 1));
	}

	if ((fabs(UB - LB) / fabs(this->MLB)) <= stop) return lb.second;

	pair <double, int> res = GreedyChoice();

	return Vec(1, res.second);
}

Vec IELearner::GetBelief()
{
	return this->b;
}


