/*
 * RTDP.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "RTDP.h"

RTDP::RTDP(Mat &Field, int N, vector <IsoGaussian*> &covfuncs, int nL, Vec &b)
{
	this->Field = Field; this->covfuncs = covfuncs;
	this->N = N; this->nL = nL; this->b = b; this->root = NULL;
	this->factory = new Factory(Field, N, covfuncs, nL);
}

Node* RTDP::Explore(Node* node, double threshold)
{
	node->Explore();
	vector <Node*> children = node->Report();

	pair <double, double> biggest = children[0]->GetBounds(), gap;

	int selected = 0;

	for (int x = 1; x < (int) children.size(); x++)
	{
		gap = children[x]->GetBounds();
		if (gap.first > biggest.first) // lower-bound first, then upper-bound
		{
			biggest = gap;
			selected = x;
		}
		else if ((fabs(gap.first - biggest.first) <= EPS) && (fabs(gap.second - gap.first) > fabs(biggest.second - biggest.first)))
		{
			biggest = gap;
			selected = x;
		}
	}

	for (int x = 0; x < (int) children.size(); x++)
		if ((x != selected) && (children[x]->Budget() > 1)) this->Q.push(children[x]);

	if ((children[selected]->Budget() > 1) && (children[selected]->BoundGap() > threshold)) return Explore(children[selected], threshold);
		else return children[selected];
}

Node* RTDP::Plan(int n, double threshold, int maxpath)
{
	root = new Node(n, 0, 0, NULL, this->factory, this->b);

	this->Q = priority_queue <Node*, vector <Node*>, NodeComparator> ();
	this->Q.push(root);

	int nPath = 0, checkpoint = 1;

	pair <double, double> oldBounds = root->GetBounds();

	while ((root->BoundGap() > threshold) && (nPath < maxpath) && (!this->Q.empty()))
	{
		Node* best = this->Q.top(); this->Q.pop();

		Node* bottom = Explore(best, threshold);

		bottom->UpdateBounds();
		nPath++;
	}

	while (!this->Q.empty()) this->Q.pop();

	return root;
}

void RTDP::UpdateData(int x, double y)
{
	Vec X(0); Vec Y(0);

	this->factory->GetPriorData(X, Y);

	vector < pair <double, double> > pGP = this->factory->PosteriorGP(X, Y, x);

	double sum = 0.0;
	Vec pb(this->nL);

	for (int t = 0; t < this->nL; t++)
	{
		pb[t] = this->b[t] * N(y, pGP[t].first, pGP[t].second);
		sum += pb[t];
	}

	for (int t = 0; t < this->nL; t++) pb[t] /= sum;

	this->b = pb;

	this->factory->UpdatePriorData(x, y);
}

Vec RTDP::GetBelief()
{
	return this->b;
}
